//  npu.c   (nat GPU)   19 Mar 2012
//		version 2 ISA   5 Dec 2012
//     	P. Chongstitvatana
//
//	    add monitor to control execution     13 Nov 2017

#include <stdlib.h>
#include "npu.h"

int R[NPE][32], LDS[NPE], PC, IR, RNG;
int M[MAXMEM];

//int op, a1, a2, a3, disp;
int runflag, stopf, icnt, cycle;

extern int display;			// display control

// instruction decoder
//  op:8 ads:14 r1:5 r2:5
#define opcode() ((IR >> 24) & 0x0ff)
#define ads()	 ((IR >> 10) & 0x03fff)
#define r1()	 ((IR >> 5) & 0x01f)
#define r2()	 (IR & 0x01f)

void error(char *mess){
	printf("error: %s\n", mess);
	exit(0);
}

// a is 14-bit, sign extend
int signx(int a){
	if( a & 0x02000 ) return a | 0xffffc000;
	return a;
}

char opsym[][8] = {
  "ld", "st", "ldr", "str", "ldx", "stx", "ldw", "bc", "add", "sub",
  "mul", "ashr", "addi", "and", "or", "xor", "lt","le","eq", "jmp",
  "jz", "jnz", "rnd", "mvt", "sys", "inc", "dec", "clr", "mov"
};

// number of cycle per instruction
int inst_cycle[] = {
	9,9,7,7,9,9,9,7,7,7,
	7,7,7,7,7,7,7,7,7,6,
	6,7,7,7,7,7,7,7,7
};

int isvalidop(int op){
	if( op < 10 || op > 38 ) return 0;
	return 1;
}

void showInst(int op, int a1, int a2, int a3){
	char opcode[8];
	if( isvalidop(op) )
		strcpy(opcode,opsym[op-10]);
	else
		strcpy(opcode,"undef");
	printf("%d: %s %d %d %d\n",PC-1,opcode,signx(a1),a2,a3);
}

// all R[k] are zero
int testrz(int k){
	int i;
	for(i = 0; i < NPE; i++)
		if( R[i][k] != 0 ) return 0;
	return 1;
}

void execute(void){
	int i, op, a1, a2, a3, disp;

	IR = M[PC]; 		// inst. fetch
	PC++;
	icnt++;
	op = opcode();		// decode bit in IR
	a1 = ads();
	disp = signx(a1);
	a2 = r1();
	a3 = r2();
	if( display ) showInst(op,a1,a2,a3);

	switch(op){
	case  xLd:  LDS[a2] = M[a1]; break;
	case  xSt:  M[a1] = LDS[a2]; break;

	// allow only one memory access from PEs
	case  xLdx:	LDS[a1] = M[R[a1][a2]+R[a1][a3]]; break;
	case  xStx:	M[R[a1][a2]+R[a1][a3]] = LDS[a1]; break;

	case  xLdr:
		for(i = 0; i < NPE; i++)
			R[i][a1] = LDS[i];
		break;
	case  xStr:
		for(i = 0; i < NPE; i++)
			LDS[i] = R[i][a1];
		break;
	case  xLdw:
		for(i = 0; i < NPE; i++)
			LDS[i] = M[a1];
		break;
	case  xBc:
		for(i = 0; i < NPE; i++)
			R[i][a1] = LDS[a2];
		break;
	case  xAdd:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] + R[i][a3];
		break;
	case  xSub:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] - R[i][a3];
		break;
	case  xMul:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] * R[i][a3];
		break;
	case  xAshr:
		for(i = 0; i < NPE; i++)
			R[i][a2] = R[i][a3] >> a1;   // unsign !
		break;
	case  xAddi:
		for(i = 0; i < NPE; i++)
			R[i][a2] = R[i][a3] + disp;  // signex
		break;
	case  xAnd:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] & R[i][a3];
		break;
	case  xOr:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] | R[i][a3];
		break;
	case  xXor:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] ^ R[i][a3];
		break;
	case  xLt:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] < R[i][a3];
		break;
	case  xLe:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] <= R[i][a3];
		break;
	case  xEq:
		for(i = 0; i < NPE; i++)
			R[i][a1] = R[i][a2] == R[i][a3];
		break;
	case  xJmp: PC = a1; break;
	case  xJz:	if( testrz(a2) ) PC = a1; break;
	case  xJnz:	if( !testrz(a2) ) PC = a1; break;
	case  xRnd: break;
	case  xMvt:
		for(i = 0; i < NPE; i++)
			if( R[i][a1] != 0 )
				R[i][a2] = R[i][a3];
		break;
	case  xSys:
		switch(a1){
		case 4: stopf = 1; break;
		}
		break;
	}
	if( isvalidop(op) )
		cycle += inst_cycle[op-10];
}

// encode fields into an instruction
int enc4(int op, int a1, int a2, int a3){
	a1 = (a1 << 10) & 0x00fffc00;	// mask 14-bit
	return (op << 24 )| a1 |(a2 << 5)| a3;
}

void loadobj(char *fname){
	int i, op, a1, a2, a3;
	int magic, start, len;
	FILE *fp;

	fp = fopen(fname, "rt");
	fscanf(fp,"%d",&magic);
	if(magic != MAGIC) error("not npu object file\n");
	fscanf(fp, "%d%d", &start, &len);
	for(i = start; i < start+len; i++){
		fscanf(fp,"%d%d%d%d",&op,&a1,&a2,&a3);
		M[i] = enc4(op,a1,a2,a3);
	}
	printf("load object to %d\n",i);
	fscanf(fp, "%d%d", &start, &len);	// data
	for(i = start; i < start+len; i++){
		fscanf(fp, "%d",&a1);
		M[i] = a1;
	}
	fclose(fp);
}

void init(void){
	PC = 0;
	runflag = 1;
	stopf = 0;
	icnt = 0;
	cycle = 0;
}

int main(int argc, char *argv[]){
	loadobj(argv[1]);
	init();
	interp();
	printf("execute %d instructions %d cycles\n",icnt,cycle);
	return 0;
}
